#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : psts.py
# Contents: functions for parallel spike trains
# Author  : Christian Borgelt
# History : 2012.07.04 file created
#           2012.07.05 function warptrain() added
#           2012.10.01 functions warppsts(), psts2mp() etc. added
#           2013.02.08 compatible with Python 3.2, added main program
#           2013.02.14 replaced "while True" with "while 1" (faster)
#           2017.05.11 specialized for pictures for fpm lecture
#-----------------------------------------------------------------------
from sys    import argv, stderr
from time   import time
from bisect import bisect
from math   import sqrt, exp, log, factorial, ceil
from random import seed, random, randint, sample
from random import expovariate, gammavariate

#-----------------------------------------------------------------------

def integrate (list):
    '''integrate (list)
Integrate a list of numbers.
list    list to integrate
returns integral as a list'''
    integ = list[0:1]           # sum elements from start of list
    for i in range(1,len(list)):
        integ.append(integ[-1]+list[i])
    return integ                # iteratively expand sum with next elem.

#-----------------------------------------------------------------------

def poissonvariate (mu):
    '''poissonvariate (mu)
Generate a Poisson variate (faster than scipy.stats.poisson.rvs()).
mu      expected value/variance of the Poisson distribution
returns a Possion variate'''
    if mu <= 0: return 0        # check for a vanishing expected value
    if mu < 10:                 # for small exp.v. use Knuth's method
        e = exp(-mu); x = 0; p = 1.0
        while 1:                # multiply uniform variates
            p *= random()       # until the result is <= e^{-mu},
            if p <= e: return x # then return the number of steps
            x += 1              # (i.e., the number of multiplications)
    b = 0.931 +2.53 *sqrt(mu); a = -0.059 +0.02483 *b
    c = log(1.1239 +1.1328 /(b-3.4))
    x = 0.9277 -3.6224 /(b-2); y = log(mu)
    while 1:                    # otherwise use rejection method
        v = random(); u = random() -0.5; z = 0.5 -abs(u)
        k = int((2*a/z + b)*u +mu +0.43)
        if z >= 0.07 and v <= x: return k
        if k < 0 or (z < 0.013 and v > z): continue
        if log(v) +c -log(a/(z*z)+b) <= k*y -mu -log(factorial(k)):
            return k
    # The above code is adapted from the numpy C implementation.
    # Using this code is more than a factor of 10 faster than calling
    # scipy.stats.poisson.rvs().

#-----------------------------------------------------------------------

def gentrain (period, rate=20, shape=1):
    '''gentrain (period, rate=20, shape=1)
Generate a single stationary spike train.
period  length of the spike train (in seconds)
rate    firing rate of the neuron        (default: 20Hz)
shape   shape parameter of Gamma process (default: 1 = Poisson)
returns a sorted list of spike times in [0,period)'''
    if rate  == 0: return []    # check for a vanishing rate
    if shape <= 0: shape = 1    # correct an illegal shape parameter
    s = int(shape)              # check for an integer shape parameter
    if shape == s and (s < 124 or rate <= 0):
        n = int(0.5-rate) if rate < 0 else poissonvariate(period*rate*s)
        t = sorted([random() for i in range(n)])
        t = [period*t[i] for i in range(randint(0,s-1),n,s)]
        return t                # generate Poisson process (and select)
    a = 1.0 /shape              # if adding time steps is needed
    b = 1.0 /rate if rate > 0 else period /float(-rate)
    x = -b *expovariate(1)      # start before zero for run-in
    while x < 0: x += a *gammavariate(shape, b)
    t = []                      # then add gamma distribution steps
    while x < period: t.append(x); x += a *gammavariate(shape, b)
    return t                    # use time points in [0,period)

#-----------------------------------------------------------------------

def warptrain (train, warp):
    '''warptrain (train, warp)
Warp a spike train.
train   spike train (list of spike times) to warp
warp    (inverse) warp function as a list (discretized function)
        Such a warp function can be computed as the integral over
        a firing rate profile, divided by the average firing rate.
        The last element in the warp list should be the length of
        the period that is warped. All spikes beyond this point
        in time are discarded.
returns a list of warped spike times (warped spike train)'''
    n   = len(warp)             # get the number of time bins
    dt  = float(warp[-1]) /n    # compute the time step size
    out = []                    # initialize the output spike train
    for t in train:             # traverse the spike times
        i = bisect(warp, t)     # find the corresponding time bin
        if   i <= 0:            # treat first bin specially
            out.append((t/warp[0]) *dt)
        elif i <  n:            # if not beyond end of warp function
            x = float(warp[i] -warp[i-1])
            out.append((i +((t-warp[i-1])/x if x != 0 else 0)) *dt)
    return out                  # transform the spike times

#-----------------------------------------------------------------------

def bintrain (train, width):
    '''bintrain (train, width)
Bin a spike train.
train   spike train (list of spike times) to bin
width   time bin width
returns a list of time bin indices containing a spike'''
    return sorted(set([int(s/width+0.5) for s in train]))

#-----------------------------------------------------------------------

def genpsts (ncnt, period, rate=20, shape=1, prob=1, jitter=0,
             groups=[], asms=[]):
    '''genpsts (ncnt, period, rate=20, shape=1, prob=1, jitter=0,
         groups=[], asms=[])
Generate parallel spike trains.
ncnt    total number of neurons
period  length of the spike trains (in seconds)
rate    default firing rate of the neurons       (default: 20Hz)
shape   default shape parameter of Gamma process (default: 1 = Poisson)
prob    default copy probability for assemblies  (default: 1)
groups  list of neuron groups with different firing rates
        Each list element must be a list or tuple, consisting of
        a list of neurons, a firing rate and an optional shape.
asms    list of neuron assembly specifications
        Each list element must be a list or tuple, consisting of
        a list of neurons, a firing rate, an optional shape and an
        optional copy probability from the mother process. The rate
        may be negative as an indicator that it must be met exactly.
returns a list of spike trains; each train is a list of spike times'''
    if shape <= 0: shape = 1.0  # correct an invalid shape parameter
    groups = [[g[0],g[1],shape] if len(g) <= 2
         else [g[0],g[1],g[2] if g[2] > 0 else 1] for g in groups]
    asms   = [[a[0],a[1],shape,prob]                   if len(a) <= 2
         else [a[0],a[1],a[2] if a[2] > 0 else 1,prob] if len(a) <= 3
         else [a[0],a[1],a[2] if a[2] > 0 else 1,a[3]] for a in asms]
    params = [[rate,shape] for i in range(ncnt)]
    for g,r,s in groups:        # set the firing rates
        for n in g:             # and the shape parameters
            params[n-1] = [r,s] # of all neurons
    for a,r,s,p in asms:        # remove coincidence firing rates
        r = (-r /float(period) if r < 0 else r) *abs(p)
        for n in a: params[n-1][0] -= r
    psts = [gentrain(period, rate if rate > 0 else 0, shape)
            for rate,shape in params]
    for asm,rate,shape,prob in asms:
        if prob == 0: continue  # traverse the neuron assemblies
        t = gentrain(period, rate, shape)
        if prob >= 1:           # generate the mother process
            for n in asm:       # traverse the assembly neurons
                x = t           # copy all coincidence spikes
                if jitter > 0: x = [s +jitter*(random()-0.5) for s in x]
                psts[n-1].extend(x) # append the coincidence spikes
        else:                   # if to copy only some of the spikes
            r = prob *len(t)    # get average number of spikes to copy
            for n in asm:       # traverse the assembly neurons
                k = int(0.5-r) if prob < 0 else poissonvariate(r)
                x = sample(t, min(k, len(t)))
                if jitter > 0: x = [s +jitter*(random()-0.5) for s in x]
                psts[n-1].extend(x) # append the copied coin. spikes
    if asms: psts = [sorted(t) for t in psts]
    return psts                 # sort and return the spike trains

#-----------------------------------------------------------------------

def warppsts (psts, warp):
    '''warppsts (psts, warp)
Warp parallel spike trains.
psts    parallel spike trains (list of lists of spike times)
warp    (inverse) warp function as a list (discretized function)
        Such a warp function can be computed as the integral over
        a firing rate profile, divided by the average firing rate.
        The last element in the warp list should be the length of
        the period that is warped. All spikes beyond this point
        in time are discarded.
returns warped parellel spike trains
        (list of lists of warped spike times)'''
    return [warptrain(t, warp) for t in psts]

#-----------------------------------------------------------------------

def binpsts (psts, width):
    '''binpsts (psts, width)
Bin parallel spike trains.
psts    parallel spike trains (list of lists of spike times)
width   time bin width
returns a list of lists of time bin indices containing a spike'''
    return [bintrain(t, width) for t in psts]

#-----------------------------------------------------------------------

def psts2tras (psts, width, period):
    '''psts2tras (psts, width, period)
Turn parallel spike trains into transactions.
psts    parallel spike trains (list of lists of spike times)
width   time bin width
period  length of the spike trains (in seconds)
returns a list of sets of neuron ids, one per time bin'''
    n    = int(ceil(period/width))
    tras = [set() for i in range(n)]
    for i in range(len(psts)):  # traverse the spike trains
        for s in psts[i]:       # assign spikes to time bins
            k = int(s/width+0.5)
            tras[k if k < n else n-1].add(i+1)
    return tras                 # return transactions

#-----------------------------------------------------------------------

def genburst (period, mode, ratio=1.25, bps=1000.0):
    '''genburst (period, mode, ratio=1.25, bps=1000.0)
Generate a burst rate profile.
period  length of the profile in seconds
mode    position of the mode, in seconds from start
ratio   ratio of the time constants
bps     time bins per second
returns a list of firing rates, one per time bin'''
    mode = float(mode)*bps      # compute the mode bin offset
    a    = mode*(ratio-1.0)/(ratio*log(ratio))
    b    = ratio*a              # compute the time constants
    n    = int(period*bps+(1.0-1e-9))       # number of time bins
    s    = 1.0/(exp(-mode/a)-exp(-mode/b))  # scaling factor
    return [s*(exp(-i/a)-exp(-i/b)) for i in range(n)]

#-----------------------------------------------------------------------

def psts2mp (out, figno, psts, period, width=100.5, height=50.3,
             frame=0.1, inset=(0.2,0.2), gap=0.1,
             pen=0.1, color=(0,0,1), labels=True, mag=None,
             xlabel='"time"', ylabel='"neurons"', name=''):
    '''psts2mp (out, figno, psts, period, width=100.5, height=50.3,
         frame=0.1, inset=(0.2,0.2), gap=0.1,
         pen=0.1, color=(0,0,1), labels=True, mag=None)
Create a dot display in Metapost format.
out     output file to write Metapost code to
psts    parallel spike trains (list of lists of spike times)
period  length of the spike trains (in seconds)
width   width  of the dot display    (in millimeters, default: 100.5)
height  height of the dot display    (in millimeters, default:  50.3)
frame   width of the frame line      (in millimeters, default: 0.15)
inset   pair of x and y inset widths (in millimeters, default: 0.4)
gap     vertical gap between rows    (in millimeters, default: 0.2)
pen     width of pen for spikes      (in millimeters, default: 0.1)
color   pen color as an RGB triple   (default: (0,0,1))'''
    sx = (width  -2*inset[0] -pen) /float(period)
    sy = (height -2*inset[1] +gap) /float(len(psts))
    dh = sy -gap
    out.write('\n')
    out.write('vardef dot (expr x, y) =\n')
    out.write('  drawdot(x*%gmm+%gmm,y*%gmm+%gmm) withcolor cdot;\n'
                      % (sx,0.5*pen,sy,0.5*dh))
    out.write('enddef;\n\n')
    out.write('beginfig(%d); %% %s\n' % (figno,name))
    out.write('  color cdot; cdot = %s;\n' % str(color))
    out.write('  pickup pensquare scaled %gmm;\n' % frame)
    out.write('  draw unitsquare xscaled %gmm yscaled %gmm' \
                      % (width, height))
    out.write(      ' shifted (-%gmm,-%gmm);\n' % inset)
    if labels:
        out.write('  save defaultscale; defaultscale := 0.5*sqrt(2);\n')
        out.write('  label.lrt(%s, (0,-0.7mm));\n'            % xlabel)
        out.write('  label.urt(%s, (0, 0.8mm)) rotated 90;\n' % ylabel)
        out.write('  pickup pensquare xscaled %gmm yscaled %gmm;\n'
                      % (pen, dh))
    out.write(' '); n = 0
    for i in range(len(psts)):
        for k in range(0,len(psts[i])):
            out.write(' dot(%f,%2d);' % (psts[i][k],i))
            n += 1
            if n % 4 == 0: out.write('\n ');
    if n % 4 != 0: out.write('\n')
    if mag:
        l,r = mag[0]; b,t = mag[1]; ox,oy = mag[2]; scale = mag[3]
        ll  = (l*sx, b*sy); w,h = (r-l)*sx, (t-b)*sy -gap
        llsrc = (l*sx-0.5*frame, b*sy    -0.5*frame)
        ursrc = (r*sx+0.5*frame, t*sy-gap+0.5*frame)
        lldst = (ll[0]        +ox-inset[0],ll[1]        +oy-inset[1])
        urdst = (ll[0]+w*scale+ox+inset[0],ll[1]+h*scale+oy+inset[1])
        s = [[x-l for x in y if x >= l and x < r] for y in psts[b:t]]
        out.write('  pickup pensquare scaled %fmm;\n' % frame)
        out.write('  draw unitsquare xscaled %fmm yscaled %fmm'
                                   % (w+frame,h+frame))
        out.write(      ' shifted (%fmm,%fmm)\n;' % llsrc)
        out.write('  path p;\n')
        out.write('  p = unitsquare xscaled %fmm yscaled %fmm'
                       % (w*scale+2*inset[0],h*scale+2*inset[1]))
        out.write(     ' shifted (%fmm,%fmm)\n;' % lldst)
        out.write('  unfill p; draw p;\n')
        out.write('  draw (%fmm,%fmm)' % llsrc);
        out.write('--(%fmm,%fmm);\n'   % lldst);
        out.write('  draw (%fmm,%fmm)' % ursrc);
        out.write('--(%fmm,%fmm);\n'   % urdst);
        out.write('  pickup pensquare xscaled %fmm yscaled %fmm;\n'
                         % (0.5*pen*scale, dh*scale))
        out.write('  vardef dot (expr x, y) =\n')
        out.write('    drawdot(x*%fmm+%fmm,y*%fmm+%fmm) withcolor cdot;\n'
                         % (sx*scale,0.5*pen*scale+ll[0]+ox,
                            sy*scale,0.5*dh *scale+ll[1]+oy))
        out.write('  enddef;\n\n')
        for i in range(len(s)):
            for t in s[i]: out.write('  dot(%f,%d);\n' % (t,i))
    out.write('endfig;\n')

#-----------------------------------------------------------------------

def burstmp (fname):
    '''burstmp (fname)
Create a MetaPost file with a dot display of a burst profile.
fname   name of the MetaPost file to write'''
    seed(7)
    n = 100; k = 12; t = 3; c = 12; cc = 20; cp = 0.75
    p = genburst(0.5,0.05)
    p = [0.0 +10 for x in range(80)] \
      + [x*60+10 for x in p[0:50]]   \
      + [x*80+30 for x in p[50:500]] \
      + [x*70+30 for x in p[0:50]]   \
      + [x*50+10 for x in p[50:420]]
    s = t/float(sum(p))
    p = [x*s for x in integrate(p)]
    refs = warppsts(genpsts(n,t), p)
    asm  = sample(range(n),k)
    print(asm)
    psts = warppsts(genpsts(n,t,asms=[[asm,-c]]), p)
    psel = warppsts(genpsts(n,t,asms=[[asm,-cc,1,cp]]), p)
    asm  = [x-1 for x in asm]
    nmap = asm +[x for x in set(range(100))-set(asm)]
    sort = [psts[nmap[i]] for i in range(len(psts))]
    psso = [psel[nmap[i]] for i in range(len(psel))]
    with open(fname, 'wt') as out:
        out.write('defaultfont  := "ptmr8r";\n')
        out.write('defaultscale := 1;\n')
        out.write('labeloffset  := 0;\n')
        psts2mp(out, 0, refs, t, name='pst_random')
        psts2mp(out, 1, refs, t, name='pst_magnify',
                mag=[(0.50,0.70),(85,93),(10,-33),8])
        psts2mp(out, 2, psts, t, name='pst_synch')
        psts2mp(out, 3, sort, t, name='pst_reord')
        psts2mp(out, 4, psel, t, name='pst_selprob')
        psts2mp(out, 5, psso, t, name='psr_spreord')
        out.write('end.\n')

#-----------------------------------------------------------------------

if __name__ == '__main__':
    burstmp('psts.mp')
